module main;
import std.stdio, std.traits, std.range, std.conv, std.functional : compose;

struct List
{
    template tmap(T) { alias T[] tmap; }
    static tmap!T unit(T)(T x) { return [x]; }
    static tmap!B fmap(A,B, alias f)(tmap!A as)
    {
        auto n = as.length;
        auto bs = new B[n]; // try n+1 here and see compile time error
        foreach(i; 0..n) bs[i] = f(as[i]);
        return bs; 
    }
}

struct Maybe
{
    struct MB(T) 
    {	
        bool hasValue = false;
        T val; 

        this(T x) { val = x; hasValue = true; } 

        bool opEquals(ref MB!T c)
        { 
            if (hasValue) return c.hasValue && val == c.val;
            else return !c.hasValue;
        }
    }

    template tmap(T) {	alias MB!T tmap;	}

    static tmap!T unit(T)(T x) { return MB!T(x); }
    static tmap!T some(T)(T x) { return unit(x); }
    static tmap!T none(T)() { return MB!T(); }
    static tmap!B fmap(A,B, alias f)(tmap!A maybex)
    {
        if (maybex.hasValue)
            return MB!B( f(maybex.val) );
            //return none!B(); // try returning and see compile time error
        else 
            return MB!B();
    }
}

//Identity:    fmap id      ≡ id
template idCheck(F)
{	
    T id(T)(T x) { return x; }	

    bool test(T)(T x) 
    {
        auto fx = F.unit!T(x);
        auto fid_fx = F.fmap!(T,T, id!T)(fx);
        return fid_fx == fx;
    }
    static if (test(1) && test("abc") && test(2.0) && test(false))
        enum idCheck = true;
    else 
        static assert(0, F.stringof ~ " identity check failed");
}

//Composition: fmap (f ∘ g) ≡ fmap f ∘ fmap g
template compCheck(F)
{
    bool g(int x) { return x < 10; }
    string f(bool x) { return to!string(x); }

    bool test(int x)
    {		
        auto fx = F.unit!int(x);
        auto left = F.fmap!(int, string, compose!(f,g))( fx );
        auto right = compose!( F.fmap!(bool, string, f), F.fmap!(int, bool, g) )(fx);
        return left == right;
    }

    enum compCheck = test(1) && test(11);
}

template isFunctor(F) {
    enum isFunctor = hasMember!(F, "unit") && hasMember!(F, "fmap") && hasMember!(F, "tmap") && idCheck!(F) && compCheck!(F);
}

void useFunctor(F)() if (isFunctor!(F))
{
    writeln(F.stringof ~ " is ok");
}

int succ(int x) { return x + 1; }

int main(string[] argv)
{
    auto lst = List.unit(5);
    auto v2 = List.fmap!(int, string, to!string)(lst);
    auto v3 = List.fmap!(int, int, succ)(lst);
    useFunctor!(List)();

    auto mb = Maybe.some!int(4);
    auto mb2 = Maybe.some!int(5);
    useFunctor!(Maybe)();
    writeln(mb == mb2);
    return 0;
}
